/*-----------------------------------------------------------------------*/
/* Low level disk I/O module skeleton for FatFs     (C)ChaN, 2013        */
/*-----------------------------------------------------------------------*/
/* If a working storage control module is available, it should be        */
/* attached to the FatFs via a glue function rather than modifying it.   */
/* This is an example of glue functions to attach various exsisting      */
/* storage control module to the FatFs module with a defined API.        */
/*-----------------------------------------------------------------------*/

#include "hardware.h"
#include "diskio.h"             /* FatFs lower layer API */

/* MMC/SD command (SPI mode) */
#define CMD0    (0)             /* GO_IDLE_STATE */
#define CMD1    (1)             /* SEND_OP_COND */
#define ACMD41  (0x80|41)       /* SEND_OP_COND (SDC) */
#define CMD8    (8)             /* SEND_IF_COND */
#define CMD9    (9)             /* SEND_CSD */
#define CMD10   (10)            /* SEND_CID */
#define CMD12   (12)            /* STOP_TRANSMISSION */
#define CMD13   (13)            /* SEND_STATUS */
#define ACMD13  (0x80|13)       /* SD_STATUS (SDC) */
#define CMD16   (16)            /* SET_BLOCKLEN */
#define CMD17   (17)            /* READ_SINGLE_BLOCK */
#define CMD18   (18)            /* READ_MULTIPLE_BLOCK */
#define CMD23   (23)            /* SET_BLOCK_COUNT */
#define ACMD23  (0x80|23)       /* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24   (24)            /* WRITE_BLOCK */
#define CMD25   (25)            /* WRITE_MULTIPLE_BLOCK */
#define CMD32   (32)            /* ERASE_ER_BLK_START */
#define CMD33   (33)            /* ERASE_ER_BLK_END */
#define CMD38   (38)            /* ERASE */
#define CMD55   (55)            /* APP_CMD */
#define CMD58   (58)            /* READ_OCR */


static
DSTATUS Stat = STA_NOINIT;      /* Disk status */

static
BYTE CardType;                  /* b0:MMC, b1:SDv1, b2:SDv2, b3:Block addressing */


/*-----------------------------------------------------------------------*/
/* Transmit bytes to the card                                            */
/*-----------------------------------------------------------------------*/

//static
void xmit_mmc (
        const BYTE* buff,       /* Data to be sent */
        UINT bc                         /* Number of bytes to send */
)
{
        BYTE d;
        do {
                d = *buff++;    /* Get a byte to be sent */
                SD_DATA = d;
        } while (--bc);
}



/*-----------------------------------------------------------------------*/
/* Receive bytes from the card (bitbanging)                              */
/*-----------------------------------------------------------------------*/

//static
void rcvr_mmc (
        BYTE *buff,     /* Pointer to read buffer */
        UINT bc         /* Number of bytes to receive */
)
{
        BYTE r;
        do {
                r = SD_DATA;
                *buff++ = r;                    /* Store a received byte */
        } while (--bc);
}

/*-----------------------------------------------------------------------*/
/* Wait for card ready                                                   */
/*-----------------------------------------------------------------------*/

static
int wait_ready (void)   /* 1:OK, 0:Timeout */
{
        BYTE d;
        UINT tmr;

        for (tmr = 25000; tmr; tmr--) { /* Wait for ready */
                rcvr_mmc(&d, 1);
                if (d == 0xFF) break;
        }

        return tmr ? 1 : 0;
}


/*-----------------------------------------------------------------------*/
/* Wait for card response                                                */
/*-----------------------------------------------------------------------*/

static
BYTE wait_response (void)
{
        UINT tmr;
        BYTE d;

        for (tmr = 250; tmr; tmr--)     /* Wait for response */
        {
                rcvr_mmc(&d,1);
                if (d != 0xFF) return d;
        }

        return d;
}


/*-----------------------------------------------------------------------*/
/* Deselect the card and release SPI bus                                 */
/*-----------------------------------------------------------------------*/

static
void deselect (void)
{
        BYTE d;
        SD_CONTROL = 0xFF;
        rcvr_mmc(&d, 1);        /* Dummy clock (force DO hi-z for multiple slave SPI) */
}



/*-----------------------------------------------------------------------*/
/* Select the card                                                       */
/*-----------------------------------------------------------------------*/

static
void select (void)
{
        BYTE d;
        SD_CONTROL = 0xFE;
        rcvr_mmc(&d, 1);        /* Dummy clock (force DO enabled) */
}



/*-----------------------------------------------------------------------*/
/* Receive a data packet from the card                                   */
/*-----------------------------------------------------------------------*/

static
int rcvr_datablock (    /* 1:OK, 0:Failed */
        BYTE *buff,                     /* Data buffer to store received data */
        UINT btr                        /* Byte count */
)
{
        BYTE d[2];
        UINT tmr;


        for (tmr = 5000; tmr; tmr--) {  /* Wait for data packet */
                rcvr_mmc(d, 1);
                if (d[0] != 0xFF) break;
        }
        if (d[0] != 0xFE) return 0;             /* If not valid data token, return with error */

        rcvr_mmc(buff, btr);                    /* Receive the data block into buffer */
        rcvr_mmc(d, 2);                                 /* Discard CRC */

        return 1;                                               /* Return with success */
}



/*-----------------------------------------------------------------------*/
/* Send a data packet to the card                                        */
/*-----------------------------------------------------------------------*/

static
int xmit_datablock (    /* 1:OK, 0:Failed */
        const BYTE *buff,       /* 512 byte data block to be transmitted */
        BYTE token                      /* Data/Stop token */
)
{
        BYTE d[2];

        d[0] = token;
        xmit_mmc(d, 1);                         /* Xmit a token */
        if (token != 0xFD) {            /* Is it data token? */
                xmit_mmc(buff, 512);    /* Xmit the 512 byte data block to MMC */

                d[0]=0xFF; d[1]=0xFF;
                xmit_mmc(d, 2);                 /* Xmit dummy CRC (0xFF,0xFF) */

                d[0] = wait_response();         /* wait for response */

                if ((d[0] & 0x1F) != 0x05)      /* If not accepted, return with error */
                        return 0;
        }

        return 1;
}



/*-----------------------------------------------------------------------*/
/* Send a command packet to the card                                     */
/*-----------------------------------------------------------------------*/

static
BYTE send_cmd (         /* Returns command response (bit7==1:Send failed)*/
        BYTE cmd,               /* Command byte */
        DWORD arg               /* Argument */
)
{
        BYTE n, d, buf[8];

        if (cmd & 0x80) {       /* ACMD<n> is the command sequense of CMD55-CMD<n> */
                cmd &= 0x7F;
                n = send_cmd(CMD55, 0);
                if (n > 1) return n;
        }

        /* Select the card */
        select();

        /* Send a command packet */
        buf[0] = 0x40 | cmd;                    /* Start + Command index */
        buf[1] = (BYTE)(arg >> 24);             /* Argument[31..24] */
        buf[2] = (BYTE)(arg >> 16);             /* Argument[23..16] */
        buf[3] = (BYTE)(arg >> 8);              /* Argument[15..8] */
        buf[4] = (BYTE)arg;                             /* Argument[7..0] */
        n = 0x01;                                               /* Dummy CRC + Stop */
        if (cmd == CMD0) n = 0x95;              /* (valid CRC for CMD0(0)) */
        if (cmd == CMD8) n = 0x87;              /* (valid CRC for CMD8(0x1AA)) */
        buf[5] = n;
        xmit_mmc(buf, 6);

        if (cmd == CMD12)
        {
                rcvr_mmc(buf, 8);       /* Skip up to 8 bytes whilst read stops */
        }

        d = wait_response();

        return d;                       /* Return with the response value */
}

/*--------------------------------------------------------------------------

   Public Functions

---------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (
        BYTE drv                        /* Drive number (always 0) */
)
{
        if (drv) return STA_NOINIT;

        return Stat;
}



/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (
        BYTE drv                /* Physical drive nmuber (0) */
)
{
        BYTE ty, cmd, buf[4];
        UINT tmr;
        DSTATUS s;


        if (drv) return RES_NOTRDY;

        /* Cancel any currently-active multiple block read */
        send_cmd(CMD12, 0);

        ty = 0;
        if (send_cmd(CMD0, 0) == 1) {                   /* Enter Idle state */
                if (send_cmd(CMD8, 0x1AA) == 1) {       /* SDv2? */
                        rcvr_mmc(buf, 4);                                                       /* Get trailing return value of R7 resp */
                        if (buf[2] == 0x01 && buf[3] == 0xAA) {         /* The card can work at vdd range of 2.7-3.6V */
                                for (tmr = 5000; tmr; tmr--) {                  /* Wait for leaving idle state (ACMD41 with HCS bit) */
                                        if (send_cmd(ACMD41, 1UL << 30) == 0) break;
                                }
                                if (tmr && send_cmd(CMD58, 0) == 0) {   /* Check CCS bit in the OCR */
                                        rcvr_mmc(buf, 4);
                                        ty = (buf[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;      /* SDv2 */
                                }
                        }
                } else {                                                        /* SDv1 or MMCv3 */
                        if (send_cmd(ACMD41, 0) <= 1)   {
                                ty = CT_SD1; cmd = ACMD41;      /* SDv1 */
                        } else {
                                ty = CT_MMC; cmd = CMD1;        /* MMCv3 */
                        }
                        for (tmr = 5000; tmr; tmr--) {                  /* Wait for leaving idle state */
                                if (send_cmd(cmd, 0) == 0) break;
                        }
                        if (!tmr || send_cmd(CMD16, 512) != 0)  /* Set R/W block length to 512 */
                                ty = 0;
                }
        }
        CardType = ty;
        s = ty ? 0 : STA_NOINIT;
        Stat = s;

        deselect();

        return s;
}



/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read (
        BYTE drv,                       /* Physical drive nmuber (0) */
        BYTE *buff,                     /* Pointer to the data buffer to store read data */
        DWORD sector,           /* Start sector number (LBA) */
        UINT count                      /* Sector count (1..128) */
)
{
        if (disk_status(drv) & STA_NOINIT) return RES_NOTRDY;
        if (!(CardType & CT_BLOCK)) sector *= 512;      /* Convert LBA to byte address if needed */

        if (count == 1) {       /* Single block read */
                if ((send_cmd(CMD17, sector) == 0)      /* READ_SINGLE_BLOCK */
                        && rcvr_datablock(buff, 512))
                        count = 0;
        }
        else {                          /* Multiple block read */
                if (send_cmd(CMD18, sector) == 0) {     /* READ_MULTIPLE_BLOCK */
                        do {
                                if (!rcvr_datablock(buff, 512)) break;
                                buff += 512;
                        } while (--count);
                        send_cmd(CMD12, 0);                             /* STOP_TRANSMISSION */
                }
        }
        deselect();

        return count ? RES_ERROR : RES_OK;
}



/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

DRESULT disk_write (
        BYTE drv,                       /* Physical drive nmuber (0) */
        const BYTE *buff,       /* Pointer to the data to be written */
        DWORD sector,           /* Start sector number (LBA) */
        UINT count                      /* Sector count (1..128) */
)
{
        BYTE d;
        DWORD lba;

        if (disk_status(drv) & STA_NOINIT) return RES_NOTRDY;

        do {
                lba = sector++;

                if (!(CardType & CT_BLOCK)) lba *= 512; /* Convert LBA to byte address if needed */

                if (send_cmd(CMD24, lba)) break;        /* WRITE_BLOCK */

                if (!xmit_datablock(buff, 0xFE)) break;

                do { ; } while (wait_response() == 0);  /* wait if card busy */

                if (send_cmd(CMD13, 0)) break;          /* SEND STATUS */

                rcvr_mmc(&d,1);         /* Get R2 from SEND STATUS */
                if (d != 0) break;      /* anything except 0 is an error */

                buff += 512;
        } while (--count);
        deselect();

        return count ? RES_ERROR : RES_OK;
}


/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

DRESULT disk_ioctl (
        BYTE drv,               /* Physical drive nmuber (0) */
        BYTE ctrl,              /* Control code */
        void *buff              /* Buffer to send/receive control data */
)
{
        DRESULT res;
        BYTE n, csd[16];
        DWORD cs;


        if (disk_status(drv) & STA_NOINIT) return RES_NOTRDY;   /* Check if card is in the socket */

        res = RES_ERROR;
        switch (ctrl) {
                case CTRL_SYNC :                /* Make sure that no pending write process */
                        select();
                        if (wait_ready()) res = RES_OK;
                        break;

                case GET_SECTOR_COUNT : /* Get number of sectors on the disk (DWORD) */
                        if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {
                                if ((csd[0] >> 6) == 1) {       /* SDC ver 2.00 */
                                        cs = csd[9] + ((WORD)csd[8] << 8) + ((DWORD)(csd[7] & 63) << 16) + 1;
                                        *(DWORD*)buff = cs << 10;
                                } else {                                        /* SDC ver 1.XX or MMC */
                                        n = (csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
                                        cs = (csd[8] >> 6) + ((WORD)csd[7] << 2) + ((WORD)(csd[6] & 3) << 10) + 1;
                                        *(DWORD*)buff = cs << (n - 9);
                                }
                                res = RES_OK;
                        }
                        break;

                case GET_BLOCK_SIZE :   /* Get erase block size in unit of sector (DWORD) */
                        *(DWORD*)buff = 128;
                        res = RES_OK;
                        break;

                default:
                        res = RES_PARERR;
        }

        deselect();

        return res;
}
